# Legobot
# Copyright (C) 2016 Brenton Briggs, Kevin McCabe, and Drew Bronson

import ssl
import threading
import logging

import irc.bot
import irc.client
import irc.connection

from Legobot.Message import Message, Metadata
from Legobot.Lego import Lego
from jaraco.stream import buffer

logger = logging.getLogger(__name__)


class IgnoreErrorsBuffer(buffer.DecodingLineBuffer):
    """  Handle char decode errors better
    """
    def handle_exception(self):
        pass


irc.client.ServerConnection.buffer_class = IgnoreErrorsBuffer
irc.client.SimpleIRCClient.buffer_class = IgnoreErrorsBuffer


class IRCBot(threading.Thread, irc.bot.SingleServerIRCBot):
    """
    Create bot instance
    """
    def __init__(self, baseplate, channels, nickname, server,
                 port=6667, use_ssl=False, password=None,
                 username=None, ircname=None, nickserv=False,
                 nickserv_pass=None):
        irc.bot.SingleServerIRCBot.__init__(self,
                                            [(server, port)],
                                            nickname,
                                            nickname)
        threading.Thread.__init__(self)

        # the obvious self.channels is already used by irc.bot
        self.my_channels = channels
        self.nickname = nickname
        self.server = server
        self.baseplate = baseplate
        self.port = port
        self.use_ssl = use_ssl
        self.password = password
        self.username = username
        self.ircname = ircname
        self.nickserv = nickserv
        self.nickserv_pass = nickserv_pass

    def connect(self, *args, **kwargs):
        """
        Connect to a server.

        This overrides the function in SimpleIRCClient
        to provide SSL functionality.

        :param args:
        :param kwargs:
        :return:
        """
        if self.use_ssl:
            factory = irc.connection.Factory(wrapper=ssl.wrap_socket)
        else:
            factory = irc.connection.Factory()
        self.connection.connect(server=self.server,
                                port=self.port,
                                nickname=self.nickname,
                                connect_factory=factory,
                                password=self.password,
                                username=self.username,
                                ircname=self.ircname)

    def on_pubmsg(self, c, e):
        """
        This function runs when the bot receives a public message.
        """
        text = e.arguments[0]
        metadata = Metadata(source=self).__dict__
        metadata['source_channel'] = e.target
        metadata['source_user'] = e.source
        metadata['source_username'] = e.source.split('!')[0]
        metadata['is_private_message'] = False
        message = Message(text=text, metadata=metadata).__dict__
        self.baseplate.tell(message)

    def on_privmsg(self, c, e):
        """
        This function runs when the bot receives a private message (query).
        """
        text = e.arguments[0]
        metadata = Metadata(source=self).__dict__
        logger.debug('{0!s}'.format(e.source))
        metadata['source_channel'] = e.source.split('!')[0]
        metadata['source_username'] = e.source.split('!')[0]
        metadata['source_user'] = e.source
        metadata['is_private_message'] = True
        message = Message(text=text, metadata=metadata).__dict__
        self.baseplate.tell(message)

    def on_welcome(self, c, e):
        """
        This function runs when the bot successfully connects to the IRC server
        """
        for channel in self.my_channels:
            logger.debug('Attempting to join {0!s}'.format(channel))
            c.join(channel)

        if self.nickserv is True and self.nickserv_pass is not None:
            self.identify(c, e, self.nickserv_pass)
        else:
            logger.error('If nickserv is enabled, you must supply a password')

        if self.nickserv is False and self.nickserv_pass is not None:
            logger.warn('It appears you provided a nickserv password but '
                        'did not enable nickserv authentication')

    def identify(self, c, e, password):
        c.privmsg('NickServ',
                  'IDENTIFY {0!s} {1!s}'.format(self.nickname, password))
        return

    def run(self):
        """
        Run the bot in a thread.

        Implementing the IRC listener as a thread allows it to
        listen without blocking IRCLego's ability to listen
        as a pykka actor.

        :return: None
        """
        self._connect()
        super(irc.bot.SingleServerIRCBot, self).start()


class IRC(Lego):

    def __init__(self, baseplate, lock, *args, **kwargs):
        super().__init__(baseplate, lock)
        self.botThread = IRCBot(baseplate, *args, **kwargs)

    def on_start(self):
        self.botThread.start()

    def listening_for(self, message):
        return str(self.botThread) != str(message['metadata']['source'])

    def handle(self, message):
        logger.info(message)
        self.botThread.connection.privmsg(message['metadata']['opts'][
            'target'], message['text'])

    def get_name(self):
        return None



# Legobot
# Copyright (C) 2016 Brenton Briggs, Kevin McCabe, and Drew Bronson

import logging
from Legobot.Lego import Lego

logger = logging.getLogger(__name__)


class Help(Lego):
    def listening_for(self, message):
        return message['text'].split()[0] == '!help'
        logger.info(message)

    def handle(self, message):
        logger.info(message)
        try:
            target = message['metadata']['source_channel']
        except IndexError:
            logger.error('Could not identify message source in message: {0!s}'
                         .format(str(message)))
        try:
            function = message['text'].split()[1]
        except IndexError:
            function = None

        baseplate_proxy = self.baseplate.proxy()
        legos = baseplate_proxy.children.get()

        help_str = 'No help is available. Sorry.'

        if not function:
            lego_names = []
            for lego in legos:
                lego_proxy = lego.proxy()
                if lego_proxy.get_name().get() is not None:
                    lego_names.append(lego_proxy.get_name().get())
            help_str = 'Available functions: ' + ', '.join(lego_names)

        if function:
            for lego in legos:
                lego_proxy = lego.proxy()
                if lego_proxy.get_name().get() == function:
                    help_str = lego_proxy.get_help().get()

        opts = {'target': target}

        self.reply(message, help_str, opts=opts)

    def get_name(self):
        return None



# Legobot
# Copyright (C) 2016 Brenton Briggs, Kevin McCabe, and Drew Bronson

import threading
import json
import logging

import pykka

from Legobot.LegoError import LegoError
from Legobot.Message import Message, Metadata


logger = logging.getLogger(__name__)


class Lego(pykka.ThreadingActor):
    class HandlerThread(threading.Thread):
        """
        This class provides a simple thread for running message handlers.
        It is used to ensure that message handlers do not block other Legos
        from running by simply taking too long to execute.
        """

        def __init__(self, handler, message):
            threading.Thread.__init__(self)
            self.handler = handler
            self.message = message

        def run(self):
            self.handler(self.message)

    def __init__(self, baseplate, lock: threading.Lock, log_file=None):
        """
        :param baseplate: the baseplate Lego, which should be \
                          the same instance of Lego for all Legos
        :param lock: a threading lock, which should be the same \
                     instance of threading.Lock for all Legos
        """
        super().__init__()
        if not lock:
            raise LegoError("Lock expected but not provided!")
        self.baseplate = baseplate
        self.children = []
        self.lock = lock
        self.log_file = log_file

    def on_receive(self, message):
        """
        Handle being informed of a message.

        This function is called whenever a Lego receives a message, as
        specified in the pykka documentation.

        Legos should not override this function.

        :param message:
        :return:
        """
        if self.log_file is not None and message['should_log']:
            message_copy = Message(message['text'],
                                   Metadata(None).__dict__,
                                   message['should_log']).__dict__
            with open(self.log_file, mode='w') as f:
                f.write(json.dumps(message_copy))
            logger.info(message['metadata']['source'])
        if self.listening_for(message):
            self_thread = self.HandlerThread(self.handle, message)
            self_thread.start()
        self.cleanup()
        for child in self.children:
            child.tell(message)

    def cleanup(self):
        """
        Clean up finished children.

        :return: None
        """
        self.lock.acquire()
        logger.debug('Acquired lock in cleanup for ' + str(self))
        self.children = [child for child in self.children if child.is_alive()]
        self.lock.release()

    def listening_for(self, message):
        """
        Return whether this Lego is listening for the provided Message.

        All Legos should override this function.

        :param message: a Message object
        :return: a boolean
        """
        return False

    def handle(self, message):
        """
        Handle the provided Message.

        All Legos should override this function.

        :param message: a Message object
        :return: None
        """
        return

    def add_child(self, child_type, *args, **kwargs):
        """
        Initialize and keep track of a child.

        :param child_type: a class inheriting from Lego to initialize \
                           an instance of
        :param args: arguments for initializing the child
        :param kwargs: keyword arguments for initializing the child
        :return:
        """
        try:
            baseplate = kwargs['baseplate']
        except:
            if self.baseplate is None:
                baseplate = self.actor_ref
            else:
                baseplate = self.baseplate
        try:
            lock = kwargs['lock']
        except:
            lock = self.lock
        child = child_type.start(baseplate, lock, *args, **kwargs)
        self.children.append(child)

    def reply(self, message, text, opts=None):
        """
        Reply to the sender of the provided message with a message \
        containing the provided text.

        :param message: the message to reply to
        :param text: the text to reply with
        :param opts: A dictionary of additional values to add to metadata
        :return: None
        """
        metadata = Metadata(source=self,
                            dest=message['metadata']['source']).__dict__
        metadata['opts'] = opts
        message = Message(text=text, metadata=metadata,
                          should_log=message['should_log']).__dict__
        self.baseplate.tell(message)

    def get_name(self):
        """
        Return the name the Lego recognizes from the help function.

        :return: a string
        """
        return '?'

    def get_help(self):
        """
        Return a helpstring for the function.

        :return: a string
        """
        return ''

    def on_failure(self, exception_type, exception_value, traceback):
        logger.exception('Lego crashed: ' + str(self))
        logger.exception(exception_type)
        logger.exception(exception_value)



'''
.. module:: Legobot.Connectors.Slack
  :synopsis: A backend for connecting Legobot to Slack

.. moduleauthor:: Bren Briggs

Module lovingly built with inspiration from slackhq's RtmBot
'''
import sys
import time
import logging
import threading

from slackclient import SlackClient

from Legobot.Message import Message, Metadata
from Legobot.Lego import Lego

logger = logging.getLogger(__name__)

sys.dont_write_bytecode = True


class RtmBot(threading.Thread, object):
    '''
    Create slack bot instance using SlackClient
    '''

    def __init__(self, baseplate, token, *args, **kwargs):
        '''
        Initialize the bot.

        :param baseplate: The baseplate/parent lego. Typically passed in from
        Legobot.Connectors.Slack.Slack

        :param token: The Slack API token you generated
        :param args: Reserved for future use
        :param kwargs: Reserved for future use
        '''
        self.baseplate = baseplate
        self.token = token
        self.last_ping = 0
        # 'event':'method'
        self.supported_events = {'message': self.on_message}
        self.slack_client = SlackClient(self.token)
        threading.Thread.__init__(self)

    def connect(self):
        '''
        Initialize connection to slack.
        :return: None
        '''

        self.slack_client.rtm_connect()

    def on_message(self, event):
        '''
        Runs when a message event is received
        :return: Legobot.Message
        '''
        metadata = self._parse_metadata(event)
        message = Message(text=event['text'],
                          metadata=metadata).__dict__
        return message

    def run(self):
        '''
        Extends the run() method of threading.Thread

        :return: None
        '''

        self.connect()
        while True:
            for event in self.slack_client.rtm_read():
                logger.debug(event)
                if event['type'] in self.supported_events:
                    event_type = event['type']
                    dispatcher = self.supported_events[event_type]
                    message = dispatcher(event)
                    logger.debug(message)
                    self.baseplate.tell(message)
            self.keepalive()
            time.sleep(0.1)
        return

    def _parse_metadata(self, message):
        '''
        Parse incoming messages to build metadata dict

        Lots of 'if' statements. It sucks, I know.

        :param message: Message dict sent from Slack

        :return: dict
        '''

        # Try to handle all the fields of events we care about.
        metadata = Metadata(source=self).__dict__
        if 'presence' in metadata:
            metadata['presence'] = message['presence']
        if 'text' in metadata:
            metadata['text'] = message['text']

        if 'user' in message:
            metadata['source_user'] = message['user']
        elif 'bot_id' in message:
            metadata['source_user'] = message['bot_id']

        if 'channel' in message:
            metadata['source_channel'] = message['channel']
            # Slack starts DM channel IDs with "D"
            if message['channel'].startswith('D'):
                metadata['is_private_message'] = True
            else:
                metadata['is_private_message'] = False

        return metadata

    def keepalive(self):
        '''
        Sends a keepalive to Slack
        '''
        # hardcode the interval to 3 seconds
        now = int(time.time())
        if now > self.last_ping + 3:
            self.slack_client.server.ping()
            self.last_ping = now


class Slack(Lego):
    '''
    Lego that builds and connects Legobot.Connectors.Slack.RtmBot

    :param baseplate: baseplate/parent lego. Typically created in your bot
    script

    :param lock: thread lock created in your bot script. All legos should
    share the same lock

    :param args: args to pass into RtmBot (like token)

    :param kwargs: keyword args to pass to RtmBot
    '''

    def __init__(self, baseplate, lock, *args, **kwargs):
        super().__init__(baseplate, lock)
        self.botThread = RtmBot(baseplate, *args, **kwargs)

    def on_start(self):
        '''
        Extends pykka's on_start method to launch this as an actor
        '''

        self.botThread.start()

    def listening_for(self, message):
        '''
        Describe what this should listen for (hint: everything)
        Extends Legobot.Lego.listening_for()
        '''

        return str(self.botThread) != str(message['metadata']['source'])

    def handle(self, message):
        '''
        Describe how this lego should handle messages.
        Extends Legobot.Lego.handle()
        '''

        logger.info(message)
        self.botThread.slack_client.api_call(
            "chat.postMessage",
            channel=message['metadata']['opts']['target'],
            text=message['text']
        )

    def get_name(self):
        '''
        Called by built-in !help lego

        Returns name of Lego. Returns none because this is
        a non-interactive Lego

        :return: None
        '''

        return None



from setuptools import setup, find_packages

setup(
    name='Legobot',

    version='1.1.1',

    license='GPLv2',

    py_modules=['Legobot'],

    description="A framework for creating interactive chatbots on various"
    "protocols",

    author="Kevin McCabe, Bren Briggs, and Drew Bronson",

    url="https://github.com/bbriggs/Legobot",

    classifiers=[
        'Development Status :: 5 - Production/Stable',

        # Pick your license as you wish (should match "license" above)
         'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3'
    ],

    packages=find_packages(exclude=['contrib', 'docs']),

    install_requires=['pykka', 'irc', 'slackclient'],
)
